package D2P::Parser;

use	Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(@allObjects @customCode @enumerations @groups %groupsOrdered $currentLine %lines %classes %classesName $currentClass);


###                      ###
#### XML PARSER PACKAGE ####
###                      ###

# Parser debugging
$DEBUGXML	= 0;

# Generic parsing
$prefix		= '';
$xmlLevel	= 0;
@allObjects	= ();
@customCode	= ();
@enumerations	= ();

# Group handling
$inGroup	= 0;
$subGroups	= 0;
$noGroup	= 0;
@groups		= ();
%groupsOrdered	= ();
$inLPackage	= 0;

# Relations and inheritance
$currentLine	= '';
%lines		= ();
$inLine		= 0;
$inLineType	= 0;
$inLConnections	= 0;

# Class handling
%classes	= ();
%classesName	= ();
$currentClass	= '';
$currentField	= '';
$currentIndex	= '';
$inClass	= 0;
$inClassName	= 0;
$inClassSte	= 0;
$inClassAttrs	= 0;
$inClassField	= 0;
$inClassOps	= 0;
$inClassIndex	= 0;

# Table field handling
$inCFName	= 0;
$inCFType	= 0;
$inCFOptions	= 0;
$inCFValue	= 0;

# Table index handling
$inCIName	= 0;
$inCIType	= 0;
$inCIUniqueness	= 0;
$inCIParams	= 0;
$inCIField	= 0;
$inCIFieldName	= 0;

# Text handling
$inStdText	= 0;
$totalText	= '';


###
## Called when parser enters a tag
#
sub	StartTag
{
	my $p = shift;				# parser context
	$ctag = shift;				# name of this tag element
	%attr = %_;				# hash with attributes

	$DEBUGXML && print STDERR (" "x($xmlLevel += 1)) . "> $ctag\n";

	# Handle diagram start for dia >= 0.83.0
	if	($ctag eq $prefix.'dia:diagram')
	{
		$prefix = 'dia:';
	}
	# Handle groups
	elsif	($ctag eq $prefix.'group')
	{
		if	($inGroup)
		{
			$subGroups ++;
			$DEBUG && print STDERR "Entering subgroup level $subGroup for group $noGroup\n";
		}
		else
		{
			$DEBUG && print STDERR "Entering top-level group $noGroup\n";
			$inGroup = 1;
			$groups[$noGroup] = {
				to		=> 'group',
				type		=> '',		# Either 'box' or 'pack'
				topLevelText	=> '',
				order		=> 0,
				classes		=> []
			};
			push @allObjects, $groups[$noGroup];
		}
	}
	# Handle standard text
	elsif	($inStdText == 1 && $ctag eq $prefix.'attribute' && $attr{name} eq 'text')
	{
		$DEBUG && print STDERR " > found 'text' attribute\n";
		$inStdText = 2;
	}
	elsif	($inStdText == 2 && $ctag eq $prefix.'string')
	{
		$DEBUG && print STDERR "  > found string\n";
		$inStdText = 3;
		$totalText = '';
	}
	elsif	($inStdText)
	{
		return;
	}
	# Package name lookup
	elsif	($inLPackage == 1 && $ctag eq $prefix.'attribute' && $attr{name} eq 'name')
	{
		$inLPackage++;
	}
	elsif	($inLPackage == 2 && $ctag eq $prefix.'string')
	{
		$inLPackage++;
		$totalText = '';
	}
	# Class name lookup
	elsif	($inClassName == 1 && $ctag eq $prefix.'string')
	{
		$inClassName++;
		$totalText = '';
	}
	# Class stereotype lookup
	elsif	($inClassSte == 1 && $ctag eq $prefix.'string')
	{
		$inClassSte++;
		$totalText = '';
	}
	# Table field name lookup
	elsif	($inCFName == 1 && $ctag eq $prefix.'string')
	{
		$inCFName++;
		$totalText = '';
	}
	# Table field type lookup
	elsif	($inCFType == 1 && $ctag eq $prefix.'string')
	{
		$inCFType++;
		$totalText = '';
	}
	# Table field options lookup
	elsif	($inCFOptions == 1 && $ctag eq $prefix.'enum')
	{
		if	($attr{val} eq '0')
		{
			$currentField->{mode} = 1;
			$currentField->{not_null} = 1;
		}
		elsif	($attr{val} eq '2')
			{ $currentField->{mode} = 2; }
		$inCFOptions = 0;
	}
	# Table field default value lookup
	elsif	($inCFValue == 1 && $ctag eq $prefix.'string')
	{
		$inCFValue++;
		$totalText = '';
	}
	# Table field 'NOT NULL' option
	elsif	($inCFNotNull == 1 && $ctag eq $prefix.'boolean')
	{
		$inCFNotNull = 0;
		if	($attr{val} eq 'true')
		{
			$currentField->{not_null} = 1;
		}
	}
	# Handle table fields
	elsif	($inClassField && $ctag eq $prefix.'attribute')
	{
		$attrLevel++;
		if	($attr{name} eq 'name')
			{ $inCFName = 1; }
		elsif	($attr{name} eq 'type')
			{ $inCFType = 1; }
		elsif	($attr{name} eq 'visibility')
			{ $inCFOptions = 1; }
		elsif	($attr{name} eq 'value')
			{ $inCFValue = 1; }
		elsif	($attr{name} eq 'class_scope')
			{ $inCFNotNull = 1; }
	}
	# Table index name lookup
	elsif	($inCIName == 1 && $ctag eq $prefix.'string')
	{
		$inCIName++;
		$totalText = '';
	}
	# Table index type lookup
	elsif	($inCIType == 1 && $ctag eq $prefix.'string')
	{
		$inCIType++;
		$totalText = '';
	}
	# Table index uniqueness check
	elsif	($inCIUniqueness == 1 && $ctag eq $prefix.'enum')
	{
		if	($attr{val} eq '1')
		{
			$currentIndex->{unique} = 1;
			$DEBUG && print STDERR " > Current index is an unique index\n";
		}
		$inCIUniqueness = 0;
	}
	# Table index fields
	elsif	($inCIFieldName == 1 && $ctag eq $prefix.'string')
	{
		$inCIFieldName++;
		$totalText = '';
	}
	elsif	($inCIField == 1 && $ctag eq $prefix.'attribute')
	{
		$attrLevel++;
		$inCIFieldName = 1 if $attr{name} eq 'name';
	}
	elsif	($inCIParams == 1 && $ctag eq $prefix.'composite' && $attr{type} eq 'umlparameter')
	{
		$inCIField = 1;
	}
	# Handle table indices
	elsif	($inClassIndex && $ctag eq $prefix.'attribute')
	{
		$attrLevel++;
		if	($attr{name} eq 'name')
			{ $inCIName = 1; }
		elsif	($attr{name} eq 'type')
			{ $inCIType = 1; }
		elsif	($attr{name} eq 'visibility')
			{ $inCIUniqueness = 1; }
		elsif	($attr{name} eq 'parameters')
			{ $inCIParams = 1; }
	}
	# Handle tables
	elsif	($inClass)
	{
		if	($ctag eq $prefix.'attribute' && $attr{name} eq 'name' && !$attrLevel)
		{
			$inClassName = 1;
			$attrLevel++;
		}
		elsif	($ctag eq $prefix.'attribute' && $attr{name} eq 'stereotype' && !$attrLevel)
		{
			$inClassSte = 1;
			$attrLevel++;
		}
		elsif	($ctag eq $prefix.'attribute' && $attr{name} eq 'attributes' && !$attrLevel)
		{
			$DEBUG && print STDERR " > start of attributes list\n";
			$inClassAttrs = 1;
		}
		elsif	($ctag eq $prefix.'composite' && $attr{type} eq 'umlattribute' && $inClassAttrs)
		{
			$DEBUG && print STDERR "  > start of attribute\n";
			$inClassField = 1;
			push (@{$currentClass->{fields}}, ($currentField = {
				to		=> 'field',
				parent		=> $currentClass,
				name		=> '',
				type		=> '',
				mode		=> 0,			# 1 = primary key
									# 2 = foreign key
				not_null	=> 0,
				references	=> '',
				default		=> ''
			}));
			push @allObjects, $currentField;
		}
		elsif	($ctag eq $prefix.'attribute' && $attr{name} eq 'operations' && !$attrLevel)
		{
			$DEBUG && print STDERR " > start of operations list\n";
			$inClassOps = 1;
		}
		elsif	($ctag eq $prefix.'composite' && $attr{type} eq 'umloperation' && $inClassOps)
		{
			$DEBUG && print STDERR "  > start of index\n";
			$inClassIndex = 1;
			push (@{$currentClass->{indices}}, ($currentIndex = {
				to		=> 'index',
				parent		=> $currentClass,
				unique		=> 0,			# [UNIQUE]
				type		=> '',			# USING type
				name		=> '',
				fields		=> []
			}));
			push @allObjects, $currentIndex;
		}
		elsif	($ctag eq $prefix.'attribute')
		{
			$attrLevel++;
		}
	}
	# Handle lines
	elsif	($inLConnections && $ctag eq $prefix.'connection' && $inLine == 1)
	{
		my	$h = {
			class	=> $attr{to},
			dest	=> $attr{connection}
		};
		if	($attr{handle} eq '1')
			{ $currentLine->{to} = $h; }
		elsif	($attr{handle} eq '0')
			{ $currentLine->{from} = $h; }
	}
	elsif	($inLConnections && $ctag eq $prefix.'connection')
	{
		my	$h = {
			class	=> $attr{to},
			dest	=> $attr{connection}
		};
		if	($attr{handle} eq '0')
			{ $currentLine->{to} = $h; }
		elsif	($attr{handle} eq '1')
			{ $currentLine->{from} = $h; }
	}
	elsif	($inLineType && $ctag eq $prefix.'enum')
	{
		if	($attr{val} eq '1')
		{
			$currentLine->{type} = 0;
		}
		elsif	($attr{val} eq '6')
		{
			$currentLine->{type} = 1;
		}
		elsif	($attr{val} eq '7')
		{
			$currentLine->{type} = 2;
		}
		elsif	($attr{val} eq '12')
		{
			$currentLine->{type} = 3;
		}
		elsif	($attr{val} eq '2')
		{
			$currentLine->{type} = 4;
		}
	}
	elsif	($inLine)
	{
		if	($ctag eq $prefix.'attribute' && $attr{name} eq 'end_arrow' && $inLine == 1)
		{
			$inLineType = 1;
		}
		elsif	($ctag eq $prefix.'connections')
		{
			$inLConnections = 1;
		}
	}
	# Handle objects
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'Standard - Text'
		&&	$inGroup == 1 && $subGroups == 0
		)
	{
		$DEBUG && print STDERR "Found standard text in top-level group $noGroup\n";
		$inStdText = 1;
	}
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'Standard - Box'
		&&	($groups[$noGroup]->{type} eq '')
		&&	$inGroup == 1 && $subGroups == 0
		)
	{
		$DEBUG && print STDERR "Found box in top-level group $noGroup\n";
		$groups[$noGroup]->{type} = 'box';
	}
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'UML - LargePackage'
		&&	($groups[$noGroup]->{type} eq '')
		&&	$inGroup == 1 && $subGroups == 0
		)
	{
		$DEBUG && print STDERR "Found UML large package in top-level group $noGroup\n";
		$groups[$noGroup]->{type} = 'pack';
		$inLPackage = 1;
	}
	elsif	(	$ctag eq $prefix.'object' && $attr{type} eq 'UML - Class'
		&&	$inGroup == 1
		)
	{
		$DEBUG && print STDERR "Found UML class in top-level group $noGroup\n";
		push (@{$groups[$noGroup]->{classes}}, ($currentClass = {
			to	=> 'class',
			dia_oid	=> $attr{id},
			parent	=> $groups[$noGroup],
			name	=> '',
			order	=> 0,
			fields	=> [],
			inherits=> [],
			indices	=> []
		}));
		$classes{$attr{id}} = $currentClass;
		push @allObjects, $currentClass;
		$inClass = 1;
		$attrLevel = 0;
	}
	elsif	(	$ctag eq $prefix.'object'
		&&	(	$attr{type} eq 'Standard - Line'
			||	$attr{type} eq 'Standard - ZigZagLine'
			)
		)
	{
		$DEBUG && print STDERR "Found line\n";
		push (@allObjects, ($currentLine = {
			to	=> 'relation',
			dia_oid	=> $attr{id},
			type	=> undef,	# 0 = foreign key
						# 1 = foreign key, ON DELETE CASCADE
						# 2 = foreign key, ON UPDATE CASCADE
						# 3 = foreign key, ON UPDATE CASCADE, ON DELETE CASCADE
						# 4 = inheritance
			from	=> undef,
			to	=> undef
		}));
		$lines{$attr{id}} = $currentLine;
		$inLine = 1;
	}
	elsif	($ctag eq $prefix.'object' && $attr{type} eq 'UML - Generalization')
	{
		$DEBUG && print STDERR "Found UML generalization\n";
		push (@allObjects, ($currentLine = {
			to	=> 'relation',
			dia_oid	=> $attr{id},
			type	=> 4,
			from	=> undef,
			to	=> undef
		}));
		$lines{$attr{id}} = $currentLine;
		$inLine = 2;
	}
}


###
## Called when parser leaves a tag
#
sub	EndTag
{
	
	my $p = shift;
	$ctag = shift;

	# Handle end of groups
	if	($ctag eq $prefix.'group')
	{
		if	($subGroups)
		{
			$DEBUG && print STDERR "Leaving subgroup level $subGroup for group $noGroup\n";
			$subGroups --;
		}
		else
		{
			$DEBUG && print STDERR "Leaving top-level group $noGroup\n";
			$inGroup = 0;
			$noGroup++;
		}
	}
	# Handle end of standard text
	elsif	($inStdText == 3 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		if	($inGroup && !$subGroups && $groups[$noGroup]->{topLevelText} eq '')
		{
			$totalText =~ /^([^\(]+)\(([0-9]+)\)$/;
			my	$name	= $1;
			my	$order	= $2;
			$name = $totalText unless $name;
			$order = 0 unless $order;
			$DEBUG && print STDERR "-> totalText = $totalText ; name = $name ; order = $order\n";

			$groups[$noGroup]->{topLevelText} = $name;			
			$groups[$noGroup]->{order} = $order;
			$groupsOrdered{$order} = [] unless $groupsOrdered{$order};
			push @{$groupsOrdered{$order}}, $groups[$noGroup];
			$DEBUG && print STDERR "   > Got top level text for group $noGroup : $totalText\n";
		}
		$inStdText = 1;
	}
	elsif	($inStdText == 1 && $ctag eq $prefix.'object')
	{
		$inStdText = 0;
	}
	# Handle large package name lookup
	elsif	($inLPackage == 3 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		if	($inGroup && !$subGroups)
		{
			($groups[$noGroup]->{topLevelText} ne '') || ($groups[$noGroup]->{topLevelText} = $totalText);
			$DEBUG && print STDERR " > Got name for large package inside top-level group $noGroup : $totalText\n";
		}
		$inLPackage = 1;
	}
	# Class name lookup
	elsif	($inClassName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentClass->{name} = $totalText;
		$classesName{$totalText} = $currentClass;
		$DEBUG && print STDERR "Current class name = $totalText\n";
		$inClassName = 0;
	}
	# Class stereotype lookup
	elsif	($inClassSte == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentClass->{order} = $totalText;
		$DEBUG && print STDERR "Current class stereotype = $totalText\n";
		$inClassSte = 0;
	}
	# Table field name lookup
	elsif	($inCFName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentField->{name} = $totalText;
		$DEBUG && print STDERR "Current field name = $totalText\n";
		$inCFName = 0;
	}
	# Table field type lookup
	elsif	($inCFType == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentField->{type} = $totalText;
		$DEBUG && print STDERR "Current field type = $totalText\n";
		$inCFType = 0;
	}
	# Table field default value lookup
	elsif	($inCFValue == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentField->{default} = $totalText;
		$DEBUG && print STDERR "Current field default = $totalText\n";
		$inCFValue = 0;
	}
	# Table fields
	elsif	($inClassField && $ctag eq $prefix.'composite')
	{
		if	($currentField->{mode} == 1)
		{
			$currentClass->{pkey} = ($currentClass->{pkey} ? $currentClass->{pkey} . "," : "") . $currentField->{name};
		}
		$inClassField = 0;
		$DEBUG && print STDERR "  > end of attribute\n";
	}
	elsif	(!$inClassField && $inClassAttrs && $ctag eq $prefix.'attribute')
	{
		$DEBUG && print STDERR " > end of attributes list\n";
		$inClassAttrs = 0;
	}
	# Table index name lookup
	elsif	($inCIName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentIndex->{name} = $totalText;
		$DEBUG && print STDERR "Current index name = $totalText\n";
		$inCIName = 0;
	}
	# Table index type lookup
	elsif	($inCIType == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		$currentIndex->{type} = $totalText;
		$DEBUG && print STDERR "Current index type = $totalText\n";
		$inCIType = 0;
	}
	# Table index field name lookup
	elsif	($inCIFieldName == 2 && $ctag eq $prefix.'string')
	{
		$totalText =~ s/\#$//;
		push @{$currentIndex->{fields}}, $totalText;
		$DEBUG && print STDERR "New index field = $totalText\n";
		$inCIFieldName = 0;
	}
	# Table index fields list
	elsif	($inCIField && $ctag eq $prefix.'composite')
	{
		$inCIField = $inCIFieldName = 0;
	}
	elsif	(!$inCIField && $inCIParams && $ctag eq $prefix.'attribute')
	{
		$inCIParams = 0;
	}
	# Table indices
	elsif	($inClassIndex && $ctag eq $prefix.'composite')
	{
		$inClassIndex = 0;
		if	(!$currentIndex->{name})
		{
			$currentIndex->{name} = "i_" . $currentIndex->{type} . "_";
			$currentIndex->{name} .= $currentIndex->{parent}->{name} . "_";
			$currentIndex->{name} .= join('_', @{$currentIndex->{fields}});
			$DEBUG && print STDERR "New index name = " . $currentIndex->{name} . "\n";
		}
		$DEBUG && print STDERR "  > end of index\n";
	}
	elsif	(!$inClassIndex && $inClassOps && $ctag eq $prefix.'attribute')
	{
		$DEBUG && print STDERR " > end of operations list\n";
		$inClassOps = 0;
	}
	elsif	($inClass && $attrLevel && $ctag eq $prefix.'attribute')
	{
		$attrLevel --;
	}
	# Relation type
	elsif	($inLineType && $ctag eq $prefix.'attribute')
	{
		$inLineType = 0;
	}
	# Handle end of line connections
	elsif	($inLConnections && $ctag eq $prefix.'connections')
	{
		$inLConnections = 0;
	}
	# Handle end of objects
	elsif	($ctag eq $prefix.'object')
	{
		if	($inClass)
		{
			$inClass = 0;
		}
		elsif	($inLine)
		{
			$inLine = 0;
		}
		elsif	($inLPackage)
		{
			$inLPackage = 0;
		}
	}

	$DEBUGXML && print STDERR (" "x($xmlLevel --)) . "> $ctag\n";
}


###
## Called for text between any tags
#
sub	Text
{
	my	$text	= $_;

	if	($totalText eq '')
	{
		$totalText = $text;
		$totalText =~ s/^\#//;
	}
	else
		{ $totalText .= $text; }
}


1;
